//
// cgmin.c
//
// Minimize the configurations by Conjugate Gradient method.
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cgmin.h"

static float maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
        (maxarg1) : (maxarg2))

#define FM_ITMAX     2000                  /* for find_minimum             */
#define FM_EPS       1.0e-12               /* for find_minimum             */
#define FM_TOL       1.0e-12               /* for find_minimum             */
 
#define LM_TOL       1.0e-8                /* no smaller than the square   */
                                           /* root of the machine accuracy */
 
#define BRAK_GOLD 1.618034
#define BRAK_GLIMIT 100.0
#define BRAK_TINY 1.0e-20
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
 
#define BRENT_ITMAX 100
#define BRENT_ZEPS 1.0e-10
#define MOV3(a,b,c, d,e,f) (a)=(d);(b)=(e);(c)=(f);
 
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/***************************************************************************/

REAL       *g_fm_vect_1;   /* temporary vector of length g_nconf */
REAL       *g_fm_vect_2;   /* temporary vector of length g_nconf */
WORD        g_nconf;       /* dimension of configuration space   */
WORD        g_num;         /* number of particles */

/***************************************************************************/

REAL *ralloc (DWORD  size)
{
      REAL  *address;
 
      if ((address = (REAL *) malloc (sizeof (REAL) * size)) == NULL)
      {
          printf ("Error allocating %ld Bytes!!\n", sizeof (REAL) * size);
          exit (0);
      }
 
      return (address);
}
 
/***************************************************************************/

REAL  lin_energy (REAL *y,
                  REAL *f,
    	          REAL  gamma,
		  REAL *lin_force)
{
    DWORD i;
    WORD  n;
    REAL  F;
    REAL  pot;

    for (i = 0; i < g_nconf; i++)
	g_fm_vect_1[i] = y[i] + gamma * f[i];

    pot = pot_force(g_fm_vect_1, g_fm_vect_2);

    for (i = 0, F = 0.0; i < g_nconf; i++)
	F -= g_fm_vect_2[i] * f[i];

    *lin_force = F;

    return pot;
}

/***************************************************************************/

void mnbrak (REAL  *y,
	     REAL  *f,
	     REAL  *ax,
	     REAL  *bx,
	     REAL  *cx,
	     REAL  *fa,
	     REAL  *fb,
	     REAL  *fc)
{
	REAL ulim,u,r,q,fu,dum,lf;

	*fa=lin_energy(y, f, *ax, &lf);
	*fb=lin_energy(y, f, *bx, &lf);
	if (*fb > *fa) {
		SHFT(dum,*ax,*bx,dum)
		SHFT(dum,*fb,*fa,dum)
	}
	*cx=(*bx)+BRAK_GOLD*(*bx-*ax);
	*fc=lin_energy(y, f, *cx, &lf);
	while (*fb > *fc) {
		r=(*bx-*ax)*(*fb-*fc);
		q=(*bx-*cx)*(*fb-*fa);
		u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
			(2.0*SIGN(FMAX(fabs(q-r),BRAK_TINY),q-r));
		ulim=(*bx)+BRAK_GLIMIT*(*cx-*bx);
		if ((*bx-u)*(u-*cx) > 0.0) {
			fu=lin_energy(y, f, u, &lf);
			if (fu < *fc) {
				*ax=(*bx);
				*bx=u;
				*fa=(*fb);
				*fb=fu;
				return;
			} else if (fu > *fb) {
				*cx=u;
				*fc=fu;
				return;
			}
			u=(*cx)+BRAK_GOLD*(*cx-*bx);
			fu=lin_energy(y, f, u, &lf);
		} else if ((*cx-u)*(u-ulim) > 0.0) {
			fu=lin_energy(y, f, u, &lf);
			if (fu < *fc) {
				SHFT(*bx,*cx,u,*cx+BRAK_GOLD*(*cx-*bx))
				SHFT(*fb,*fc,fu,lin_energy(y, f, u, &lf))
			}
		} else if ((u-ulim)*(ulim-*cx) >= 0.0) {
			u=ulim;
			fu=lin_energy(y, f, u, &lf);
		} else {
			u=(*cx)+BRAK_GOLD*(*cx-*bx);
			fu=lin_energy (y, f, u, &lf);
		}
		SHFT(*ax,*bx,*cx,u)
		SHFT(*fa,*fb,*fc,fu)
	}
}

/***************************************************************************/

REAL dbrent (REAL  *y,
	     REAL  *f,
	     REAL   ax,
	     REAL   bx,
	     REAL   cx,
	     REAL   tol,
	     REAL  *xmin)
{
	int iter,ok1,ok2;
	REAL a,b,d,d1,d2,du,dv,dw,dx,e=0.0;
	REAL fu,fv,fw,fx,olde,tol1,tol2,u,u1,u2,v,w,x,xm,lf;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=lin_energy(y, f, x, &dx);
	dw=dv=dx;
	for (iter=1;iter<=BRENT_ITMAX;iter++) {
		xm=0.5*(a+b);
		tol1=tol*fabs(x)+BRENT_ZEPS;
		tol2=2.0*tol1;
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			return fx;
		}
		if (fabs(e) > tol1) {
			d1=2.0*(b-a);
			d2=d1;
			if (dw != dx) d1=(w-x)*dx/(dx-dw);
			if (dv != dx) d2=(v-x)*dx/(dx-dv);
			u1=x+d1;
			u2=x+d2;
			ok1 = (a-u1)*(u1-b) > 0.0 && dx*d1 <= 0.0;
			ok2 = (a-u2)*(u2-b) > 0.0 && dx*d2 <= 0.0;
			olde=e;
			e=d;
			if (ok1 || ok2) {
				if (ok1 && ok2)
					d=(fabs(d1) < fabs(d2) ? d1 : d2);
				else if (ok1)
					d=d1;
				else
					d=d2;
				if (fabs(d) <= fabs(0.5*olde)) {
					u=x+d;
					if (u-a < tol2 || b-u < tol2)
						d=SIGN(tol1,xm-x);
				} else {
					d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
				}
			} else {
				d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
			}
		} else {
			d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
		}
		if (fabs(d) >= tol1) {
			u=x+d;
			fu=lin_energy(y, f, u, &lf);
		} else {
			u=x+SIGN(tol1,d);
			fu=lin_energy(y, f, u, &lf);
			if (fu > fx) {
				*xmin=x;
				return fx;
			}
		}
		lin_energy (y, f, u, &du);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			MOV3(v,fv,dv, w,fw,dw)
			MOV3(w,fw,dw, x,fx,dx)
			MOV3(x,fx,dx, u,fu,du)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				MOV3(v,fv,dv, w,fw,dw)
				MOV3(w,fw,dw, u,fu,du)
			} else if (fu < fv || v == x || v == w) {
				MOV3(v,fv,dv, u,fu,du)
			}
		}
	}

	printf ("DBRENT> Too many iterations!\n");
	return 0.0;
}

/***************************************************************************/

void line_minimum (REAL *y,
	 	   REAL *f,
		   REAL *S)
{
    REAL xx, xmin, fx, fb, fa, bx, ax;
    WORD i;

    /* initial guess for bracketing */
    ax = (REAL) 0.0;
    xx = (REAL) 0.001;

    /* Bracketing */
    mnbrak (y, f, &ax, &xx, &bx, &fa, &fx, &fb);

    /* Use Brent's method to find minimum */
    *S = dbrent (y, f, ax, xx, bx, LM_TOL, &xmin);

    /* Construct the position of the minimum */
    for (i = 0; i < g_nconf; i++)
    {
	f[i] *= xmin;
	y[i] += f[i];
    }
}

/***************************************************************************/

void find_minimum (REAL  *y,
                   REAL  *ymin,
		   REAL  *Smin,
		   REAL  *grad,
		   DWORD *niter)
{
    /* Local variables */
    REAL  *g;
    REAL  *h;
    REAL  *x;
    REAL   S, Slin;
    REAL   gg, dgg, gamma;
    DWORD  i, j, k;
    WORD  n;
    REAL   zz;

    /* Allocate vectors */
    g = ralloc (g_nconf);
    h = ralloc (g_nconf);
    x = ralloc (g_nconf);

    /* local downhill gradient */
    S = pot_force (y, x); 

    /* Initialize vectors */
    for (j = 0; j < g_nconf; j++)
    {
	g[j]    = x[j];
	h[j]    = x[j];
	ymin[j] = y[j];
    }

    gg = 0.0;

    /* Loop over iterations */
    for (i = 0; i < FM_ITMAX; i++)
    {
	/* line minimization */
	line_minimum (ymin, x, &Slin);

	if( i==0 ) printf("Slin=%f\n",Slin);

	/* if minimum is found */
        if (2.0 * fabs (Slin - S) <= FM_TOL * (fabs(Slin) + fabs(S) + FM_EPS))
	    break;

        /* New action and gradient */
	S = pot_force (ymin, x);

	/* Calculate products (Polak-Ribiere method) */
	for (j = 0, gg = 0.0, dgg = 0.0; j < g_nconf; j++)
	{
	   gg  += g[j] * g[j];
	   dgg += (x[j] - g[j]) * x[j];
	}

	/* if gradient == 0 */
	if (gg == 0.0)
	    break;

        /* Compute gamma */
	gamma = dgg / gg;

	/* Compute new G and H */
	for (j = 0; j < g_nconf; j++)
	{
	    g[j] = x[j];
	    h[j] = g[j] + gamma * h[j];
	    x[j] = h[j];
	}
    }

    /* Warning if too many iterations */
    if (i == FM_ITMAX)
        printf ("FIND_MAXIMUM> More than %d  iterations!\n", FM_ITMAX);

    /* return values */
    *grad  = sqrt (gg);
    *niter = i;
    *Smin  = Slin;

    /* Free vectors */
    free (g);
    free (h);
    free (x);
}

/***************************************************************************/

int main(int argc, char *argv[])
{
  REAL *y, *ymin;
  FILE *fp;
  char s[100];
  WORD i;
  REAL r1,r2,r3;
  REAL Smin, grad;
  DWORD niter;
  REAL len, size;

  if(argc<3)
  {
    printf( "Usage: %s input_xyz_file output_xyz_file\n", argv[0] );
    return(1);
  }

  if( (fp=fopen(argv[1], "r"))==NULL )
  {
    printf( "Can't open file %s for reading", argv[1] );
    return(2);
  }

  //
  // Read configuration
  //
  fscanf( fp, "%d\n", &g_num );
  g_nconf = g_num*3;
  g_fm_vect_1  = ralloc (g_nconf);
  g_fm_vect_2  = ralloc (g_nconf);
  y = ralloc(g_nconf);
  ymin = ralloc(g_nconf);

  fgets( s, 100, fp );

  for( i=0; i<g_num; i++ )
  {
    fscanf( fp, "%s %lf %lf %lf\n", s, &r1, &r2, &r3 );
    y[i*3] = r1;   y[i*3+1] = r2;   y[i*3+2] = r3;
  }
  
  fclose(fp);

  //
  // Initialize the cell
  //
  len = 0.0;
  for( i=0; i<g_nconf; i++ )
  {
    double d;
    d = fabs(y[i]);
    if( d > len ) len = d;
  }

  size = 3.90000001;
  cellAlloc( (int)ceil(1.5*len/size), size);

  //
  // Find the minimum structure by iteraction
  //
  find_minimum( y, ymin, &Smin, &grad, &niter );

  printf("Smin=%f\n",Smin);

  //
  // Write optimized configuration to file
  //
  if( (fp=fopen(argv[2],"w")) == NULL )
  {
    printf( "Can't open file %s for writing", argv[2] );
    return(2);
  }

  fprintf( fp,"%d\n",g_num );
  fprintf( fp,"Minimized structure\n" );

  for( i=0; i<g_num; i++ )
  {
    fprintf( fp, "Au %f %f %f\n", ymin[i*3], ymin[i*3+1], ymin[i*3+2] );
  }

  fclose(fp);

  //
  // Free memories
  //
  free(g_fm_vect_1);
  free(g_fm_vect_2);
  free(y);   free(ymin);
  cellFree();

  return 0;
}
