//
// force.c
//
// Written by:  Yanting Wang    Jun. 10, 2002
//
// Gold glue potential and force used by cgmin.c.
//
// Please refer to 
// http://www.sissa.it/~furio/potentials/Au/ for glue potential
// http://www.sissa.it/furio/md/md/node21.html for Velocity Verlet algorithm
//

#include <time.h>
#include <stdlib.h>

#include "cgmin.h"

extern g_num;
extern g_nconf;
extern p1,p2;

static long m_pn=0;
static double *dx=NULL,*dy=NULL,*dz=NULL,*r=NULL;
static long *pp1=NULL,*pp2=NULL;

REAL pot_force(REAL *y, REAL *g)
{
  REAL Ep = 0.0;
  DWORD i,j,k,pn;
  REAL *dens,*du,rh;
  REAL fx,fy,fz,f;

  dens = ralloc(g_num);
  du = ralloc(g_num);

  for( i=0; i<g_num; i++ ) dens[i] = 0.0;
  for( i=0; i<g_nconf; i++ ) g[i] = 0.0;

  cellUpdate(y);

  //
  // prepare the various values for all particles
  //
  pn = 0;
  while( cellGen() )
  {
    pn++;
    if( pn>m_pn )
    {
      m_pn = pn;
      dx = (double *)realloc( dx, pn*sizeof(double) );
      dy = (double *)realloc( dy, pn*sizeof(double) );
      dz = (double *)realloc( dz, pn*sizeof(double) );
      r  = (double *)realloc( r,  pn*sizeof(double) );
      pp1 = (long *)realloc( pp1, pn*sizeof(long) );
      pp2 = (long *)realloc( pp2, pn*sizeof(long) ); 
    }

    k = pn-1;
    pp1[k] = i = p1;   pp2[k] = j = p2;
    dx[k] = y[i*3]   - y[j*3];
    dy[k] = y[i*3+1] - y[j*3+1];
    dz[k] = y[i*3+2] - y[j*3+2];
    r[k] = sqrt( dx[k]*dx[k] + dy[k]*dy[k] + dz[k]*dz[k] );

    rh = rho( r[k] );
    dens[i] += rh;
    dens[j] += rh;
  }

  for( i=0; i<g_num; i++ ) 
  {
    Ep += u( dens[i] );
    du[i] = -dU( dens[i] );
  }

  //
  // pair and glue potentials
  //

  for( k=0; k<pn; k++ )
  {
    i = pp1[k];   j = pp2[k];

    f = ( (du[i]+du[j])*dRho(r[k]) - dPhi(r[k]) ) / r[k];
    
    fx = f * dx[k]; 
    fy = f * dy[k]; 
    fz = f * dz[k];

    g[i*3]   += fx;     g[j*3]   -= fx;
    g[i*3+1] += fy;     g[j*3+1] -= fy;
    g[i*3+2] += fz;     g[j*3+2] -= fz;
    
    Ep += phi( r[k] );
  }

  free(dens);
  free(du);

  return Ep;
}

const double RRB = 3.5;
const double RRC = 3.9;
const double RRD = 2.878207442141723;
const double RHOD = 1.0;
const double DB = 12.0;
const double DSW = 9.358157767784574;
const double UB = -3.3;
const double RC = 3.7;
const double A = 4.0704;
const double D0 = 2.878207442141723;

const double R1I = -0.68;
const double R2I = 0.75;
const double R3I = -1.333333333333333;
const double R1II = -0.68;
const double R2II = 0.75;
const double R3II = -1.527241171296038;
const double R2III = 5.578188675490974;
const double R3III = 6.132971688727435;

const double A0I = -0.08;
const double A1I = 0.0;
const double A2I = 0.7619231375231362;
const double A3I = -0.8333333333333333;
const double A4I = -0.1211483464993159;
const double A0II = -0.08;
const double A1II = 0.0;
const double A2II = 0.7619231375231362;
const double A3II = -0.8333333333333333;
const double A4II = -1.096009851140349;
const double A5II = 2.158417178555998;
const double A6II = -0.9128915709636862;
const double A3III = 0.0;
const double A4III = 0.0;
const double A5III = 0.0;

const double B0I = -2.793388616771698;
const double B1I = -0.3419999999999999;
const double B2I = 0.03902327808424106;
const double B3I = 0.007558829951858879;
const double B4I = 0.0003090472511796849;
const double B2II = 0.08618226772941980;
const double B3II = 0.004341701445034724;
const double B4II = -0.0003044398779375916;
const double B2III = 0.08618226772941980;
const double B3III = 0.004325981467602070;

REAL dRho( REAL r )
{
  REAL x;

  if( r >= RRC ) return 0.0;   // after cut radius it is all zero
  else if( r >= RRB )   // region III ( release ) with a spline
  {
    x = r - RRC;
    return x * ( 2.0 * R2III + x * 3.0 * R3III );
  }
  else if( r >= RRD )   // region II ( sustain ) with a spline
  {
    x = r - RRD;
    return R1II + x * ( 2.0 * R2II + x * 3.0 * R3II );
  }
  else   // region I ( decay ) with a spline
  {
    x = r - RRD;
    return R1I + x * ( 2.0 * R2I + x * 3.0 * R3I );
  }
}

REAL rho( REAL r )
{
  REAL x;

  if( r >= RRC ) return 0.0;   // after cut radius it is all zero
  else if( r >= RRB )   // region III ( release ) with a spline
  {
    x = r - RRC;
    return x * x * ( R2III + x * R3III );
  }
  else if( r >= RRD )   // region II ( sustain ) with a spline
  {
    x = r - RRD;
    return RHOD + x * ( R1II + x * ( R2II + x * R3II ) );
  }
  else   // region I ( decay ) with a spline
  {
    x = r - RRD;
    return RHOD + x * ( R1I + x * ( R2I + x * R3I ) );
  }
}

REAL dU( REAL dens )
{
  REAL x;

  if( dens > DB )   // region III
  {
    x = dens - DB;
    return x * ( 2.0 * B2III + x * 3.0 * B3III );
  }
  else if( dens > DSW )   // region II
  {
    x = dens - DB;
    return x * ( 2.0 * B2II + x * ( 3.0 * B3II + x * 4.0 * B4II ) );
  }
  else   // region I
  {
    x = dens - DSW;
    return B1I + x * ( 2.0 * B2I + x * ( 3.0 * B3I + x * 4.0 * B4I ) );
  }
}

REAL u( REAL dens )
{
  REAL x;

  if( dens > DB )   // region III
  {
    x = dens - DB;
    return UB + x * x * ( B2III + x * B3III );
  }
  else if( dens > DSW )   // region II
  {
    x = dens - DB;
    return UB + x * x * ( B2II + x * ( B3II + x * B4II ) );
  }
  else   // region I
  {
    x = dens - DSW;
    return B0I + x * ( B1I + x * ( B2I + x * ( B3I + x * B4I ) ) );
  }
}

REAL dPhi( REAL r )
{
  REAL x;

  if( r >= RC ) return 0.0;
  else if ( r >= A )   // region III, after second neighbours
  {
    x = r - RC;
    return x * x * ( 5.0 * A5III * x * x + 4.0 * A4III * x + 3.0 * A3III );
  }
  else if( r >= D0 )   // region II, between first and second neighbours
  {
    x = r - D0;
    return A1II + x * ( 2.0 * A2II + x * ( 3.0 * A3II +
	   x * ( 4.0 * A4II + x * ( 5.0 * A5II + x * 6.0 * A6II ) )));
  }
  else   // region I, before first neighbours
  {
    x = r - D0;
    return A1I + x * ( 2.0 * A2I + x * ( 3.0 * A3I + x * 4.0 * A4I ) );
  }
}

REAL phi( REAL r )
{
  REAL x;

  if( r >= RC ) return 0.0;
  else if ( r >= A )   // region III, after second neighbours
  {
    x = r - RC;
    return x * x * x * ( A5III * x * x + A4III * x + A3III );
  }
  else if( r >= D0 )   // region II, between first and second neighbours
  {
    x = r - D0;
    return A0II + x * ( A1II + x * ( A2II + x * ( A3II +
	   x * ( A4II + x * ( A5II + x * A6II ) ))) );
  }
  else   // region I, before first neighbours
  {
    x = r - D0;
    return A0I + x * ( A1I + x * ( A2I + x * ( A3I + x * A4I ) ) );
  }
}
